/**
  Draw a CPU-side 2D image, with a 3D camera.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2012-04-25 (Public Domain)
*/
#include "physics/world.h" /* physics::library and physics::object */
#include "ogl/glsl.h"

/* Just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"


#include "osl/glsl.h" /* C++ version of GLSL primitives */
using glsl::fract;
using glsl::mod;
using glsl::reflect;
using glsl::refract;
double current_time;
#include "raytracer.cpp"


/// Make a new empty OpenGL texture with this size and internal format.
GLuint make_tex(GLenum texFormat) {
	GLuint tex;
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	/* turn off mipmapping */
	GLenum filt=GL_LINEAR; // _MIPMAP_LINEAR;
	if (texFormat==GL_RGBA_FLOAT16_ATI) filt=GL_NEAREST; /* or else nVidia 6200 fails below */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filt);
	return tex;
}

unsigned char to_byte(float v) {
	if (v<0.0) return 0;
	else if (v>1.0) return 255;
	else return (unsigned char)(255.9*v);
}

/* Convert vec3 to an RGBA byte value */
void draw_pixel(const vec3 &color,unsigned char *rgba) 
{
	rgba[0]=to_byte(color.x);
	rgba[1]=to_byte(color.y);
	rgba[2]=to_byte(color.z);
	rgba[3]=255;
}

/* Convert this vector out of this coordinate frame */
vec3 apply(const ortho_frame &orient,const vec3 &v)
{
	return orient.x*v.x+orient.y*v.y+orient.z*v.z;
}

/* A raytraced object */
class rayObject : public physics::object {
public:
	rayObject(void) 
		:physics::object(0.01)  /* <- our timestep, in seconds */
	{ }
	
	void simulate(physics::library &lib) { }
	
	void draw(physics::library &lib) {
		int wid=lib.screenw, ht=lib.screenh;
		int subsample=4; /* CPU rendering is really slow.  Subsampling speeds it up. */
		if (lib.key_toggle['l']) subsample=16; /* low resolution */
		if (lib.key_toggle['h']) subsample=1; /* high resolution (pixel for pixel) */
		wid/=subsample; ht/=subsample;
		
		unsigned char *rgba=new unsigned char[4*wid*ht];
		current_time=lib.time;
		float aspect=ht/(float)wid;
	#pragma omp parallel for schedule(dynamic,4)
		for (int y=0;y<ht;y++)
		for (int x=0;x<wid;x++) 
		{
			vec3 dir_screen=normalize(vec3(x*(2.0/wid)-1.0, aspect*(y*(2.0/ht)-1.0), -1.0));
			vec3 dir_world=apply(slow_orient,dir_screen);
			vec3 color=ray_to_color(camera,dir_world);
			draw_pixel(color,&rgba[4*(y*wid+x)]);
		}
		
		GLuint tex=make_tex(GL_RGBA8);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, 
			wid,ht, 0, GL_RGBA, GL_UNSIGNED_BYTE, rgba);
		delete[] rgba;
		
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glEnable(GL_TEXTURE_2D);
		glVertex4f(1.0,1.0,1.0,1.0);
		glBegin(GL_TRIANGLE_FAN);
		glTexCoord2f(0,0); glVertex2f(-1,-1);
		glTexCoord2f(1,0); glVertex2f(+1,-1);
		glTexCoord2f(1,1); glVertex2f(+1,+1);
		glTexCoord2f(0,1); glVertex2f(-1,+1);
		glEnd();
		glDisable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, 0);
		
		glDeleteTextures(1,&tex);
	}
};


/* Called to create a new simulation */
void physics_setup(physics::library &lib) {
	lib.world->add(new rayObject());
	lib.screenw=lib.screenh=100;
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* Point and line smoothing needs usual alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	lib.world->draw(lib);
}


