
/* Raytracer framework */
const float invalid_t=1.0e9; // far away
const float close_t=1.0e-5; // too close (behind head, self-intersection, etc)

/* This struct describes how a surface looks */
struct surface_hit_t {
	float shiny; /* 0: totally matte surface; 1: big highlight */
	vec3 reflectance; /* diffuse color */
	float mirror; /* proportion of perfect mirror specular reflection */
	float refract; /* if not 1.0, the refractive index of the material */
	
	surface_hit_t() {}
	surface_hit_t(float s,vec3 r,float m,float f) {
		shiny=s; reflectance=r; mirror=m; refract=f;
	}
};

/* This struct describes everything we know about a ray-object hit. */
struct ray_hit_t {
	vec3 P; /* world coords location of hit */
	vec3 N; /* surface normal of the hit */
	float t; /* ray t value at hit (or invalid_t if a miss) */
	float opacity; /* fraction of incoming light that makes it through us */
	surface_hit_t s;
};

vec3 calc_world_color(vec3 C,vec3 D);

/* Return the t value where the ray (C,D) hits 
    the sphere with this center and radius r. */
void sphere_hit(ray_hit_t & rh,vec3 C,vec3 D,   // ray parameters
		vec3 center,float r, // object parameters
		surface_hit_t surface)  // shading parameters
{
	float a=dot(D,D);
	float b=2.0*dot(C-center,D);
	float c=dot(C-center,C-center)-r*r;
	float det=b*b-4.0*a*c;
	if (det<0.0) return; /* miss: can't take square root of negative */
	float t=(-b-sqrt(det))/(2.0*a);
	if (t<close_t || t>rh.t) return; /* behind head or another object */
	
	vec3 P=C+t*D; // ray-object intersection point (world coordinates)
	
	/* If we got here, we're the closest hit so far. */
	rh.t=t;
	rh.P=P;
	rh.N=normalize(P-center);
	rh.s=surface;
	
	if (rh.s.mirror>0.0) 
	{ // add 70's style mirror tiles
		//rh.s.mirror*=abs(cos(10.0*P.x)*sin(10.0*P.y));
	}
}
/* Return the t value where the ray (C,D) hits 
    the disk with normal N and radius r. */
void disk_hit(ray_hit_t &rh,vec3 C,vec3 D,   // ray parameters
		vec3 N,vec3 center,float r, // object parameters
		surface_hit_t surface)  // shading parameters
{
	float a=dot(C-center,N);
	float b=dot(D,N);
	float t=-a/b;
	if (t<close_t || t>rh.t) return; // we don't see this object.
	
	vec3 P=C+t*D; // ray-object intersection point (world coordinates)
	float Pr=length(P-center)/r;
	if (Pr>1.0) return; // trim surface
	
	/* If we got here, we're the closest hit so far. */
	rh.t=t; // +0.1*sin(10.0*P.y);
	rh.P=P;
	rh.N=N;
	rh.s=surface;
	
	if (rh.s.mirror>0.0) 
	{ // add 70's style mirror tiles
		rh.s.mirror*=fabs(cos(10.0*P.x)*sin(10.0*P.y));
	}
	
	if (rh.s.refract!=1.0) { /* add waves! */
		float time=current_time;
		float speedx=2.0, speedx2=2.5, speedy=3.0;
		float szx=7.0, szx2=2.0, szy=9.0;
		float ax=0.07, ax2=0.1, ay=0.03;
		
		/* This is the underlying heightfield
		float h=-ax/szx*cos(rh.P.x*szx+speedx*time) 
		        -ax2/szx2*cos(rh.P.x*szx2+speedx2*time) 		
				-ay/szy*cos(rh.P.y*szy+speedy*time);
		rh.P.z+=h-(ax/szx+ay/szy);
		*/
		
		/* Approximate surface normal using heightfield's gradient:
		    normal deviation = dh/dx, dh/dy, dh/dz 
		*/
		rh.N=normalize(rh.N+vec3(
				ax*sin(rh.P.x*szx+speedx*time)
				+ax2*sin(rh.P.x*szx2+speedx2*time),
				ay*sin(rh.P.y*szy+speedy*time),
				0.0
				)
			);
		
		/* Approximate caustics using heightfield's curvature:
		  curvature = d^2 h/dx^2  +  d^2 h /dy^2 */
		float curvature=
			 ax*szx*cos(rh.P.x*szx+speedx*time)
			+ax2*szx2*cos(rh.P.x*szx2+speedx2*time)
			+ay*szy*cos(rh.P.y*szy+speedy*time);
		
		rh.opacity=0.5+0.3*curvature;
	}
}

/* Return a ray_hit for this world ray.  Tests against all objects. */
ray_hit_t world_hit(vec3 C,vec3 D)
{
	ray_hit_t rh; rh.t=invalid_t; rh.s.mirror=0.0; rh.opacity=1.0;
	
// Intersect new ray with all the world's geometry:
	// Ground plane
	disk_hit(rh,C,D, normalize(vec3(0.0,0.0,1.0)),vec3(0.0,0.0,-2.0),2000.0,
		 surface_hit_t(0.0,vec3(0.4,0.5,0.2),0.0,1.0));
		 
	// Beach
	sphere_hit(rh,C,D, vec3(-10.0,0.0,-21.0),23.0,
		 surface_hit_t(0.0,vec3(0.4,0.3,0.2),0.0,1.0));
	
	// Mirror sphere, off to right
	sphere_hit(rh,C,D, vec3(3.0,0.0,1.0+0.02*sin(3.5*current_time)),1.0,
		 surface_hit_t(1.0,vec3(0.4,0.4,0.4),0.8,1.0));
	
	// Flat refractive plane
	disk_hit(rh,C,D, normalize(vec3(0.0,0.0,1.0)),vec3(0.0,0.0,0.0),2000.0,
		 surface_hit_t(1.0,vec3(0.0,0.1,0.2),0.0,1.35));
	
	// Flat plane, with mirror
	disk_hit(rh,C,D, normalize(vec3(1.0,0.0,0.1)),vec3(0.0,0.0,1.0),1.4,
		 surface_hit_t(1.0,vec3(1.0,1.0,1.0)*0.1,0.8,1.0));
	
	return rh;
}

vec3 calc_world_color(vec3 C,vec3 D) {
	float frac=1.0; /* fraction of object light that makes it to the camera */
	vec3 color=vec3(0.0); /* totals up the final color of this pixel */
	vec3 skyColor=vec3(0.4,0.6,1.0);
	
	/* Intersect camera ray with world geometry */
	ray_hit_t rh=world_hit(C,D);

	if (rh.t>=invalid_t) { // Sky
		return skyColor*(1.5-D.z);
	}
	// else World object: do lighting
	float frontSide=1.0;
	if (dot(rh.N,D)>0.0) {
		rh.N=-rh.N; // flip normal to face right way
		rh.s.refract=1.0/rh.s.refract; // leaving the surface
		frontSide=0.0; 
	}

	vec3 L=normalize(vec3(0.8,-0.5,0.7));
	vec3 H=normalize(L+normalize(-D));
	float specular=rh.s.shiny*powf(clamp(dot(H,rh.N),0.0,1.0),5000.0);
	float diffuse=clamp(dot(rh.N,L),0.0,1.0);

	// check shadow ray 
	ray_hit_t shadow=world_hit(rh.P,L);
	if (shadow.t<invalid_t) {diffuse*=(1.0-shadow.opacity); specular=0.0;}
	float ambient=0.3;

	vec3 curObject=(ambient+diffuse)*rh.s.reflectance+specular*vec3(1.0);
	if (((C+rh.P)/2).z<=0.0) 
	{ /* under water: attenuate */
		float transmitFrac=exp(-0.1*rh.t);
		curObject=transmitFrac*curObject+(1.0-transmitFrac)*vec3(0.0,0.1,0.2);
	}

	color+=frac*curObject;

/* Check for mirror bounce */
	if (frac>0.02) { // only do more rays if needed
	  float doReflect=rh.s.mirror;
	  vec3 I=normalize(D);
	  vec3 N=normalize(rh.N);
	  if (rh.s.refract!=1.0) 
	  { /* reflection/refraction */

		/* Mirror reflection first */
		float mirror=0.08+0.92*pow(1.0-dot(-I,N),5.0); /* fresnel */
		if (frontSide) {
			doReflect+=mirror;
			frac*=(1.0-mirror); // all non-mirror light is refracted
		}

		/* Now refraction */
		vec3 R;
		float eta=1.0/rh.s.refract;
		float k = 1.0 - eta * eta * (1.0 - dot(N, I) * dot(N, I));
		if (k < 0.0) {
			doReflect=1.0; // total internal reflection 
		}
		else
		{ /* refraction */
			R = eta * I - (eta * dot(N, I) + sqrt(k)) * N;
			color+=frac*calc_world_color(rh.P,R);
		}
	  }
	  if (doReflect>0.0) 
	  { /* pure mirror reflection */
		color+=doReflect*frac*calc_world_color(rh.P,reflect(I,N));
	  }
	}
	return color;
}



// C++ version:
vec3 ray_to_color(vec3 camera,vec3 dir_world)
{
	return calc_world_color(camera,dir_world);
}

/* 
// GLSL version: 
void main(void) {
	vec3 C=camera; // origin of ray (world coords)
	vec3 D=location-camera; // direction of ray (world coords)

	gl_FragColor.rgb=calc_world_color(C,D);
	gl_FragColor.a=1.0; // opaque
}
*/
